"""Clock widget with time from the web."""

from __future__ import annotations

import asyncio
from datetime import datetime, timedelta, timezone
import logging
from typing import Optional

from httpx import AsyncClient, ConnectError, HTTPStatusError, ReadTimeout
from PySide6 import QtWidgets
from tenacity import before_sleep_log, retry, retry_if_exception_type, wait_exponential

LOGGER = logging.getLogger(__name__)


class WebClock(QtWidgets.QLabel):
    """Clock widget with time from the web."""

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize widget."""
        super().__init__(parent)
        self.setObjectName("clock")
        self.async_client = AsyncClient()
        self._last_sync = datetime.now(timezone.utc)
        self._async_tasks: list[asyncio.Task] = []
        self._sync_interval = 60

        self._async_tasks.append(asyncio.create_task(self._update_time()))

    @retry(
        retry=(
            retry_if_exception_type(HTTPStatusError)
            | retry_if_exception_type(ReadTimeout)
            | retry_if_exception_type(ConnectError)
        ),
        wait=wait_exponential(multiplier=1, min=0.15, max=2),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
    )
    async def _update_time(self) -> None:
        """Update time."""
        self._last_sync, offset = await self._get_time_offset()
        while True:
            current_time = datetime.now(timezone.utc)

            if (current_time - self._last_sync).total_seconds() > self._sync_interval:
                self._last_sync, offset = await self._get_time_offset()

            formated_time = (current_time - timedelta(seconds=offset)).strftime(
                "%H:%M:%S",
            )
            self.setText(f"{formated_time} UTC")

            await asyncio.sleep(1)

    async def _get_time_offset(self) -> tuple[datetime, float]:
        """Get time."""
        response = await self.async_client.get(
            "http://worldtimeapi.org/api/timezone/Etc/UTC",
        )
        response.raise_for_status()
        data = response.json()
        api_time = datetime.fromisoformat(data["datetime"])
        local_time = datetime.now(timezone.utc)
        offset = (local_time - api_time).total_seconds()
        LOGGER.debug("Synced time from web, current offest %s", offset)
        return local_time, offset



"""Wdiget to display PnL breakdown."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from PySide6.QtCore import (
    QEvent,
    Qt,
)
from PySide6.QtWidgets import (
    QLabel,
    QToolTip,
    QVBoxLayout,
    QWidget,
)

if TYPE_CHECKING:
    from decimal import Decimal

    from PySide6.QtGui import QEnterEvent, QMouseEvent


class PnlBreakdown(QWidget):
    """Pnl breakdown widget."""

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        """Initialize widget."""
        super().__init__(parent)
        self._tooltip_content = ""
        self._main_layout = QVBoxLayout(self)
        self._main_layout.setContentsMargins(0, 0, 0, 0)
        self.pnl_label = QLabel(self)
        self.pnl_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.pnl_label.setObjectName("pnl")
        self._main_layout.addWidget(self.pnl_label)
        self.setLayout(self._main_layout)

        self._show_tooltip = False

    def set_pnl(self, usd: Decimal, percent: Decimal) -> None:
        """Set pnl."""
        color = "green" if percent > 0 else "red"
        text_format = (
            f"<span style='color:{color}'>{round(usd, 3)} USD<br>{round(percent, 3)}%</span>"
        )
        self.pnl_label.setText(text_format)

    def set_tooltip_content(
        self,
        pnl: Decimal,
        funding_fee: Decimal,
        position_fee: Decimal,
        pnl_after_fee: Decimal,
        push_tool_tip: bool = True,
    ) -> None:
        """Set tooltip content."""
        self._tooltip_content = (
            f"PnL: {round(pnl, 3)}<br>"
            f"Funding Fee: -{round(funding_fee, 3)}<br>"
            f"Opening Fee: -{round(position_fee, 3)}<br>"
            f"Closing Fee: -{round(position_fee, 3)}<br><br>"
            f"PnL After Fees: {round(pnl_after_fee, 3)}"
        )
        if self._show_tooltip and push_tool_tip:
            QToolTip.showText(
                self.mapToGlobal(self.rect().center()),
                self._tooltip_content,
            )
        elif not push_tool_tip:
            self.setToolTip(self._tooltip_content)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Override event to show tooltip."""
        if event.button() == Qt.MouseButton.RightButton:
            self._show_tooltip = True
            return None
        return super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """Override event to hide tooltip."""
        if event.button() == Qt.MouseButton.RightButton:
            self._show_tooltip = False
            QToolTip.hideText()
            return None
        return super().mouseReleaseEvent(event)

    def enterEvent(self, event: QEnterEvent) -> None:
        """Override event to change cursor on hover."""
        self.setCursor(Qt.CursorShape.WhatsThisCursor)
        return super().enterEvent(event)

    def leaveEvent(self, event: QEvent) -> None:
        """Override event to reset cursor on leave."""
        self.setCursor(Qt.CursorShape.ArrowCursor)
        return super().leaveEvent(event)



"""News Filters."""

from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Protocol

from textsearch import TextSearch

if TYPE_CHECKING:
    from plutus_terminal.core.news.filter._actions import FilterAction
    from plutus_terminal.core.types_ import NewsData


class FilterBase(Protocol):
    """Protocol for news filters."""

    def add_to_queue(
        self,
        match_pattern: dict[str, str],
        filter_action: FilterAction,
        kwargs: dict,
    ) -> None:
        """Add filter to queue to be executed."""

    def execute(self, news_data: NewsData) -> NewsData:
        """Execute filter on newsData."""
        ...


class KeywordMatchingFilter(FilterBase):
    """Filter news by keywords.

    Go over the text on the body and quote of the news and search for
    provided matches, then execute the associated action.

    """

    def __init__(self) -> None:
        """Initialize shared variables."""
        self._actions_to_execute: dict[str, tuple[FilterAction, dict]] = OrderedDict()
        self.text_search = TextSearch(case="insensitive", returns="object")

    def add_to_queue(
        self,
        match_pattern: dict[str, str],
        filter_action: FilterAction,
        kwargs: dict[str, Any],
    ) -> None:
        """Add filter to queue to be executed.

        Args:
            match_pattern (dict[str, str]): Dict with data to match.
                * KeywordMatchingFilter expects `{"keyword": str}`
            filter_action (FilterAction): Action to execute.
            kwargs (dict): Keyword arguments.
        """
        keyword_match = match_pattern["keyword"].lower()
        self._actions_to_execute[keyword_match] = (filter_action, kwargs)

        # Add all keywords to text search when adding to queue to improve performance.
        self.text_search.add(list(self._actions_to_execute.keys()))

    def execute(self, news_data: NewsData) -> NewsData:
        """Execute filter on newsData."""
        for part in ("body", "quote"):
            search_result = self.text_search.findall(news_data[part])  # type: ignore
            # Inverse to start from the end of text to preserve index
            for result in search_result[::-1]:
                match_action, match_kwargs = self._actions_to_execute[result.match.lower()]
                news_data = match_action(news_data, part, result, **match_kwargs)

                # Return ealier if this news is being ignored.
                if news_data["ignored"]:
                    return news_data

        return news_data


class DataMatchingFilter(FilterBase):
    """Filter news by matching data.

    Go over every added key in the news data and search for provided matches, then
    execute the associated action.
    """

    def __init__(self) -> None:
        """Initialize shared variables."""
        self._actions_by_data_key: dict[str, dict[str, tuple[FilterAction, dict]]] = OrderedDict()

    def add_to_queue(
        self,
        match_pattern: dict[str, str],
        filter_action: FilterAction,
        kwargs: dict[str, Any],
    ) -> None:
        """Add filter to queue to be executed.

        Args:
            match_pattern (dict[str, str]): Dict with data to match.
                * DataMatchingFilter expects `{"keyword": str, "data_key": str}`
            filter_action (FilterAction): Action to execute.
            kwargs (dict): Keyword arguments.
        """
        keyword_match = match_pattern["keyword"].lower()
        data_key = match_pattern["data_key"].lower()
        if data_key not in self._actions_by_data_key:
            self._actions_by_data_key[data_key] = {}

        self._actions_by_data_key[data_key][keyword_match] = (
            filter_action,
            kwargs,
        )

    def execute(self, news_data: NewsData) -> NewsData:
        """Execute filter on newsData."""
        for data_key, actions in self._actions_by_data_key.items():
            data = news_data.get(data_key, None)
            if data is None:
                continue
                # TODO: Log error
                # (f"Key {key} used in filter not found in news data.")

            # Check if data is set due to how coins are stored in NewsData
            if isinstance(data, set):
                data = {item.lower() for item in data}
                for keyword_match in data.intersection(actions.keys()):
                    filter_action, kwargs = actions[keyword_match]
                    news_data = filter_action(news_data, "", None, **kwargs)

                    # Return ealier if this news is being ignored.
                    if news_data["ignored"]:
                        return news_data
            elif isinstance(data, str):
                if data.lower() in actions:
                    filter_action, kwargs = actions[data.lower()]
                    news_data = filter_action(news_data, "", None, **kwargs)

                    # Return ealier if this news is being ignored.
                    if news_data["ignored"]:
                        return news_data
        return news_data



"""Dialog to manage orders."""

from decimal import Decimal
from functools import partial
from typing import Optional

from PySide6 import QtWidgets
from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QPixmap

from plutus_terminal.core.exchange.base import ExchangeBase
from plutus_terminal.core.exchange.types import (
    OrderData,
    PerpsPosition,
    PerpsTradeDirection,
    PerpsTradeType,
)
from plutus_terminal.ui import ui_utils
from plutus_terminal.ui.widgets.decimal_spin_box import DecimalSpinBoxWithButton
from plutus_terminal.ui.widgets.pnl_breakdown import PnlBreakdown
from plutus_terminal.ui.widgets.top_bar_widget import TopBar


class ManageOrder(QtWidgets.QDialog):
    """Dialog to manage orders."""

    execute_order = Signal(OrderData)

    def __init__(
        self,
        order_data: OrderData,
        exchange: ExchangeBase,
        associated_position: Optional[PerpsPosition],
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize dialog."""
        super().__init__(parent)
        self._order_data = order_data
        self._exchange = exchange
        self._associated_position = associated_position

        self._main_layout = QtWidgets.QGridLayout()

        self.top_bar = TopBar()

        self._type_group = QtWidgets.QButtonGroup()
        self.limit_button = QtWidgets.QRadioButton("Limit")
        self.tp_button = QtWidgets.QRadioButton("Take Profit")
        self.sl_button = QtWidgets.QRadioButton("Stop Loss")
        self._type_group_layout = QtWidgets.QHBoxLayout()

        self._info_frame = QtWidgets.QFrame()
        self._open_price_label = QtWidgets.QLabel("Open Price:")
        self._open_price_value = QtWidgets.QLabel("--")
        self._liq_price_label = QtWidgets.QLabel("Est. Liq. Price:")
        self._liq_price_value = QtWidgets.QLabel("--")
        self._info_layout = QtWidgets.QGridLayout()
        self._pnl_label = QtWidgets.QLabel("Est. PnL:")
        self._pnl_value = PnlBreakdown()

        self._trigger_label = QtWidgets.QLabel("Trigger Price:")
        self.trigger_box = DecimalSpinBoxWithButton(button_text="USD")
        self.trigger_max_button = QtWidgets.QPushButton("MAX")

        self._amount_label = QtWidgets.QLabel("Amount:")
        self.amount_box = DecimalSpinBoxWithButton(button_text="USD")
        self.amount_max_button = QtWidgets.QPushButton("MAX")

        self.execute_order_button = QtWidgets.QPushButton("Execute Order")

        self._setup_widgets()
        self._setup_layout()

        self.setMinimumHeight(self.sizeHint().height())
        self.setMinimumWidth(int(self.sizeHint().width() * 1.75))

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self.setModal(False)
        self.setWindowIcon(QPixmap(":/icons/plutus_icon"))
        self.setWindowTitle("Manage Order")

        title_color = (
            "color: rgb(255, 100, 100);"
            if self._order_data["trade_direction"].value == PerpsTradeDirection.SHORT.value
            else "color: rgb(100, 255, 100);"
        )
        self.top_bar.title.setText(
            f"{self._exchange.format_simple_pair_from_pair(self._order_data['pair'])} | "
            f"<span style='{title_color};'>{self._order_data['trade_direction'].name}</span>",
        )

        minimum_digits = ui_utils.get_minimal_digits(float(self._order_data["trigger_price"]), 4)
        self._open_price_value.setText(f"${self._order_data["trigger_price"]:,.{minimum_digits}f}")
        self._open_price_value.setAlignment(Qt.AlignmentFlag.AlignRight)
        self._liq_price_value.setAlignment(Qt.AlignmentFlag.AlignRight)

        self._type_group.addButton(self.limit_button, PerpsTradeType.LIMIT.value)
        self._type_group.addButton(self.tp_button, PerpsTradeType.TRIGGER_TP.value)
        self._type_group.addButton(self.sl_button, PerpsTradeType.TRIGGER_SL.value)
        self._type_group.button(self._order_data["order_type"].value).click()

        self._type_group.button(PerpsTradeType.LIMIT.value).setDisabled(
            self._order_data["reduce_only"],
        )

        trigger_minimum_digits = ui_utils.get_minimal_digits(
            float(self._order_data["trigger_price"]),
            4,
        )
        self.trigger_box.setDecimals(trigger_minimum_digits)
        self.trigger_box.decimalValueChanged.connect(self.update_pnl)
        self.trigger_box.setValue(self._order_data["trigger_price"])

        self.trigger_max_button.setObjectName("actionButton")
        self.trigger_max_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.trigger_max_button.setFixedWidth(50)
        self.trigger_max_button.clicked.connect(
            partial(self.trigger_box.setValue, self._order_data["trigger_price"]),
        )

        self.amount_box.setDecimals(4)
        self.amount_box.setValue(self._order_data["size_stable"])
        self.amount_box.setMaximum(self._order_data["size_stable"])
        self.amount_box.decimalValueChanged.connect(self.on_quantity_change)

        self.amount_max_button.setObjectName("actionButton")
        self.amount_max_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.amount_max_button.setFixedWidth(50)
        self.amount_max_button.clicked.connect(
            partial(self.amount_box.setValue, self.amount_box.maximum()),
        )

        self._info_frame.setObjectName("newsFrameQuote")
        if not self._order_data["reduce_only"]:
            self._info_frame.hide()
        self._pnl_value.pnl_label.setAlignment(Qt.AlignmentFlag.AlignRight)

        self.execute_order_button.setProperty("class", "LONG")
        self.execute_order_button.setMinimumHeight(30)
        self.execute_order_button.clicked.connect(self.on_execute_order)

        self.update_liquidation_price()

    def _setup_layout(self) -> None:
        """Configure layouts."""
        self._type_group_layout.addWidget(self.limit_button)
        self._type_group_layout.addWidget(self.tp_button)
        self._type_group_layout.addWidget(self.sl_button)

        self._main_layout.addWidget(self.top_bar, 0, 0, 1, 3)
        self._main_layout.addLayout(self._type_group_layout, 1, 0, 1, 3)
        self._main_layout.addWidget(self._trigger_label, 2, 0)
        self._main_layout.addWidget(self.trigger_box, 2, 1)
        self._main_layout.addWidget(self.trigger_max_button, 2, 2)
        self._main_layout.addWidget(self._amount_label, 3, 0)
        self._main_layout.addWidget(self.amount_box, 3, 1)
        self._main_layout.addWidget(self.amount_max_button, 3, 2)

        self._info_layout.addWidget(self._open_price_label, 0, 0)
        self._info_layout.addWidget(self._open_price_value, 0, 1)
        self._info_layout.addWidget(self._liq_price_label, 1, 0)
        self._info_layout.addWidget(self._liq_price_value, 1, 1)
        self._info_layout.addWidget(self._pnl_label, 2, 0)
        self._info_layout.addWidget(self._pnl_value, 2, 1)
        self._info_frame.setLayout(self._info_layout)
        self._main_layout.addWidget(self._info_frame, 4, 0, 1, 3)

        self._main_layout.addWidget(self.execute_order_button, 5, 0, 1, 3)

        self.setLayout(self._main_layout)

    def update_liquidation_price(self) -> None:
        """Update liquidation price."""
        if self._associated_position is None:
            self._liq_price_value.setText("--")
            return

        liquidation_price = self._exchange.calculate_liquidation_price(self._associated_position)
        minimal_digits = ui_utils.get_minimal_digits(float(liquidation_price), 4)
        self._liq_price_value.setText(
            f"${liquidation_price:,.{minimal_digits}f}",
        )

    def update_pnl(self, price: Decimal) -> None:
        """Update pnl."""
        if self._associated_position is None:
            self._pnl_value.pnl_label.setText("--")
            return

        pnl_details = self._exchange.calculate_pnl(self._associated_position, price)

        self._pnl_value.set_pnl(
            pnl_details["pnl_usd_after_fees"],
            pnl_details["pnl_percentage_after_fees"],
        )
        self._pnl_value.set_tooltip_content(
            pnl_details["pnl_usd_before_fees"],
            pnl_details["funding_fee_usd"],
            pnl_details["position_fee_usd"],
            pnl_details["pnl_usd_after_fees"],
            push_tool_tip=False,
        )

    def on_quantity_change(self, new_quantity: Decimal) -> None:
        """Handle quantity change.

        Args:
            new_quantity (float): New quantity.
        """
        # If order is associated with a position, update position size
        if self._associated_position is not None:
            self._associated_position["position_size_stable"] = new_quantity
        self.update_pnl(self.trigger_box.value())

    def set_edit_mode(self, edit_mode: bool) -> None:
        """Set if the widget is editing or creating order.

        Args:
            edit_mode (bool): If true, the widget is in edit mode.
        """
        self.amount_box.setDisabled(edit_mode)
        self.amount_max_button.setDisabled(edit_mode)
        for button in self._type_group.buttons():
            button.setDisabled(edit_mode)

    def on_execute_order(self) -> None:
        """Handle execute order button click."""
        # Create order to execture based on current state
        type_button_id = self._type_group.id(self._type_group.checkedButton())
        order_type = PerpsTradeType(type_button_id)
        order = OrderData(
            id=self._order_data["id"],
            pair=self._order_data["pair"],
            trade_direction=self._order_data["trade_direction"],
            order_type=order_type,
            trigger_price=Decimal(self.trigger_box.value()),
            size_stable=Decimal(self.amount_box.value()),
            reduce_only=self._order_data["reduce_only"],
        )
        self.execute_order.emit(order)
        self.close()

    def show(self) -> None:
        """Override show method to update button position."""
        super().show()
        self.trigger_box.update_button_position()
        self.amount_box.update_button_position()



"""Combo box to select account."""

from __future__ import annotations

from typing import Optional

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import QComboBox, QWidget

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.db.models import KeyringAccount
from plutus_terminal.ui.widgets.new_account import NewAccountDialog


class AccountPicker(QComboBox):
    """Combo box to select account."""

    account_changed = Signal(KeyringAccount)

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)
        self._current_index = 0

        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        self._add_all_accounts()
        self._set_current_account()
        self.setMinimumWidth(self.sizeHint().width() + 5)
        self.currentIndexChanged.connect(self._on_account_changed)

    def _add_all_accounts(self) -> None:
        """Add all accounts."""
        self.clear()
        all_accounts = CONFIG.get_all_keyring_accounts()
        # Add all accounts
        for account in all_accounts:
            icon = QPixmap(f":/exchanges/{account.exchange_name}")
            self.addItem(
                icon,
                str(account.username).replace("_", " ").capitalize(),
                userData=account,
            )

        # Add custom item to add new accounts
        self.addItem(
            QPixmap(":icons/user_add"),
            "Add new account",
            userData="New Account",
        )

    def _set_current_account(self) -> None:
        """Set current account."""
        for index in range(self.count()):
            if self.itemData(index) == CONFIG.current_keyring_account:
                self.setCurrentIndex(index)
                self._current_index = index
                break

    def _on_account_changed(self, index: int) -> None:
        """Account changed."""
        account = self.itemData(index)

        if account == "New Account":
            new_account_dialog = NewAccountDialog()

            if not new_account_dialog.exec():
                self.blockSignals(True)
                self.setCurrentIndex(self._current_index)
                self.blockSignals(False)
                return

            account = new_account_dialog.new_account
            if not account:
                return

            self.blockSignals(True)
            self._add_all_accounts()
            self._set_current_account()
            self.blockSignals(False)
            self.account_changed.emit(account)
            return

        CONFIG.current_keyring_account = account
        self.account_changed.emit(account)



"""Web3 Provider that cicles between mutiple providers to avoid rate limiting."""

import asyncio
import itertools
import logging
from typing import Any

from aiohttp import ClientResponseError
import orjson
from web3 import AsyncHTTPProvider, AsyncWeb3
from web3.providers.async_base import AsyncJSONBaseProvider
from web3.types import RPCEndpoint, RPCResponse

from plutus_terminal.core.config import CONFIG

LOGGER = logging.getLogger(__name__)


class AsyncCycleWeb3Provider(AsyncJSONBaseProvider):
    """Web3 Provider that cicles between mutiple providers to avoid rate limiting."""

    def __init__(self, providers: list[AsyncHTTPProvider]) -> None:
        """Initialize web3 provider.

        Args:
            providers (list[AsyncWeb3]): Web3 providers.
        """
        self._providers = providers
        self._providers_cycle = itertools.cycle(providers)
        self._lock = asyncio.Lock()
        super().__init__()

    async def make_request(self, method: RPCEndpoint, params: Any) -> RPCResponse:  # noqa: ANN401
        """Make request to provider.

        Args:
            method (str): Method to call.
            params (list): Params to pass.

        Returns:
            dict: Response.
        """
        async with self._lock:
            provider = next(self._providers_cycle)
        try:
            result = await provider.make_request(method, params)
        except ClientResponseError:
            LOGGER.warning("Provider %s is not available. Removing it...", provider)
            async with self._lock:
                self._providers.remove(provider)
                self._providers_cycle = itertools.cycle(self._providers)
            return await self.make_request(method, params)

        return result


def build_cycle_provider(chain_name: str) -> AsyncWeb3:
    """Build cycle web3 provider using Web3RPC from database.

    Args:
        chain_name (str): Chain name.

    Returns:
        AsyncCycleWeb3Provider: Web3 provider.
    """
    providers_urls = orjson.loads(str(CONFIG.get_web3_rpc_by_name(chain_name).rpc_urls))
    providers = [AsyncHTTPProvider(url) for url in providers_urls]

    provider = AsyncCycleWeb3Provider(providers)

    return AsyncWeb3(provider)
